from django.db import models

from sage_encrypt.mixins.cast import INT, BIGINT, DATE, TIMESTAMP, BOOL, DOUBLE, TIME, NUMERIC
from sage_encrypt.mixins.encrypt import EncryptSymmetricMixin
from sage_encrypt.mixins.query import SYM_ENCRYPT_SQL_WITH_NULLIF


class EncryptedEmailField(EncryptSymmetricMixin, models.EmailField):
    pass


class EncryptedIntegerField(EncryptSymmetricMixin, models.IntegerField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = INT

    def get_internal_type(self):
        return 'IntegerField'


class EncryptedBigIntegerField(EncryptSymmetricMixin, models.BigIntegerField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = BIGINT

    def get_internal_type(self):
        return 'BigIntegerField'


class EncryptedTextField(EncryptSymmetricMixin, models.TextField):
    pass


class EncryptedCharField(EncryptSymmetricMixin, models.CharField):
    pass


class EncryptedDateField(EncryptSymmetricMixin, models.DateField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = DATE

    def get_internal_type(self):
        return 'DateField'


class EncryptedDateTimeField(EncryptSymmetricMixin, models.DateTimeField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = TIMESTAMP

    def get_internal_type(self):
        return 'DateTimeField'


class EncryptedBooleanField(EncryptSymmetricMixin, models.BooleanField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = BOOL

    def get_internal_type(self):
        return 'BooleanField'


class EncryptedFloatField(EncryptSymmetricMixin, models.FloatField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = DOUBLE

    def get_internal_type(self):
        return 'FloatField'


class EncryptedTimeField(EncryptSymmetricMixin, models.TimeField):
    encrypt_query = SYM_ENCRYPT_SQL_WITH_NULLIF
    cast = TIME

    def get_internal_type(self):
        return 'TimeField'

class EncryptedDecimalField(EncryptSymmetricMixin, models.DecimalField):
    cast = NUMERIC

    def get_cast_sql(self):
        return self.cast % {
            'max_digits': self.max_digits,
            'decimal_places': self.decimal_places
        }

    def get_internal_type(self):
        return 'DecimalField'



from sage_encrypt.mixins.cast import TEXT
from sage_encrypt.mixins.query import (
    SYM_ENCRYPT_SQL,
    SYM_DECRYPT_SQL,
    ASYM_ENCRYPT_SQL,
    ASYM_DECRYPT_SQL
)
from sage_encrypt.mixins.sql import DecryptedCol

from django.utils.functional import cached_property
from django.db.models.sql.compiler import SQLInsertCompiler
from django.db.backends.postgresql.base import DatabaseWrapper

from sage_encrypt.services.setting import get_setting


class Encrypt:
    """
    PostgreSQL pgcrypto encryption for django fields
    """
    encrypt_query = None
    decrypt_query = None
    cast = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def db_type(self, connection: DatabaseWrapper):
        """
        value stored in database is hexadecimal
        :param connection: connection object
        :type connection: DatabaseWrapper
        :return: db type
        :rtype: str
        """
        return 'bytea'

    def get_placeholder(self, value, compiler: SQLInsertCompiler, connection: DatabaseWrapper):
        """
        tell postgres to encrypt this field using pgcrypto
        :return: encrypt sql query
        :rtype: str
        """
        raise NotImplementedError('The `get_placeholder` needs to be implemented.')

    def get_cast_sql(self):
        """
        this may be overridden by some implementations
        :return: cast type
        :rtype: str
        """
        return self.cast

    def get_decrypt_sql(self, connection: DatabaseWrapper):
        """
        get decrypt sql query
        :param connection: connection object
        :type connection: DatabaseWrapper
        :return: decrypt sql query
        :rtype: str
        """
        raise NotImplementedError('The `get_decrypt_sql` needs to be implemented.')

    def get_internal_type(self):
        """
        may be overridden by some implementations
        :return: field type
        :rtype: str
        """
        raise NotImplementedError('The `get_internal_type` needs to be implemented.')

    def get_col(self, alias, output_field=None):
        """
        get the decryption for col
        :return: db col
        :rtype: DecryptedCol
        """
        if output_field is None:
            output_field = self
        if alias != self.model._meta.db_table or output_field != self:
            return DecryptedCol(
                alias,
                self,
                output_field
            )
        else:
            return self.cached_col

    @cached_property
    def cached_col(self):
        """
        get cached version of decryption for col
        """
        return DecryptedCol(
            self.model._meta.db_table,
            self
        )

class EncryptSymmetricMixin(Encrypt):
    """
    pgcrypto symmetric key encrypted field mixin
    """
    encrypt_query = SYM_ENCRYPT_SQL
    decrypt_query = SYM_DECRYPT_SQL
    cast = TEXT

    def get_placeholder(self, value, compiler: SQLInsertCompiler, connection: DatabaseWrapper):
        """
        set encrypt key
        :return: encrypt sql query
        :rtype: str
        """
        return self.encrypt_query.format(get_setting(connection, 'ENCRYPT_KEY'))

    def get_decrypt_sql(self, connection: DatabaseWrapper):
        """
        set decrypt key
        :param connection: connection object
        :type connection: DatabaseWrapper
        :return: decrypt sql query
        :rtype: str
        """
        return self.decrypt_query.format(get_setting(connection, 'ENCRYPT_KEY'))

    def get_internal_type(self):
        """
        may be overridden by some implementations
        :return: field type
        :rtype: str
        """
        return 'BinaryField'

class EncryptAsymmetricMixin(Encrypt):
    """
    pgcrypto asymmetric key encrypted field mixin
    """
    encrypt_query = ASYM_ENCRYPT_SQL
    decrypt_query = ASYM_DECRYPT_SQL
    cast = TEXT

    def get_placeholder(self, value, compiler: SQLInsertCompiler, connection: DatabaseWrapper):
        """
        set encrypt key
        :return: encrypt sql query
        :rtype: str
        """
        return self.encrypt_query.format(get_setting(connection, 'ENCRYPT_PUBLIC_KEY'))

    def get_decrypt_sql(self, connection: DatabaseWrapper):
        """
        set decrypt key
        :param connection: connection object
        :type connection: DatabaseWrapper
        :return: decrypt sql query
        :rtype: str
        """
        return self.decrypt_query.format(get_setting(connection, 'ENCRYPT_PRIVATE_KEY'))

    def get_internal_type(self):
        """
        may be overridden by some implementations
        :return: field type
        :rtype: str
        """
        return 'BinaryField'



from sage_encrypt import settings

def get_setting(connection, key):
    """
    get key from connection or default to settings.
    :type connection
    :type key: str
    :rtype setting value
    """
    if key in connection.settings_dict:
        return connection.settings_dict[key]
    else:
        return getattr(settings, key)



from django.db.models.expressions import Col
from django.db.models.sql.compiler import SQLInsertCompiler
from django.db.backends.postgresql.base import DatabaseWrapper


class DecryptedCol(Col):
    def __init__(self, alias, target, output_field=None):
        self.target = target
        super(DecryptedCol, self).__init__(alias, target, output_field)

    def as_sql(self, compiler: SQLInsertCompiler, connection: DatabaseWrapper):
        """
        create SQL query with PostgreSQL decryption and casting
        """
        sql, params = super(DecryptedCol, self).as_sql(compiler, connection)
        sql = self.target.get_decrypt_sql(connection) % (sql, self.target.get_cast_sql())
        return sql, params



ASYM_ENCRYPT_SQL_WITH_NULLIF = "pgp_pub_encrypt(nullif(%s, NULL)::text, dearmor('{}'))"
SYM_ENCRYPT_SQL_WITH_NULLIF = "pgp_sym_encrypt(nullif(%s, NULL)::text, '{}')"

ASYM_ENCRYPT_SQL = "pgp_pub_encrypt(%s, dearmor('{}'))"
SYM_ENCRYPT_SQL = "pgp_sym_encrypt(%s, '{}')"

ASYM_DECRYPT_SQL = "pgp_pub_decrypt(%s::bytea, dearmor('{}'))::%s"
SYM_DECRYPT_SQL = "pgp_sym_decrypt(%s::bytea, '{}')::%s"



from django.conf import settings

ENCRYPT_KEY = settings.ENCRYPT_KEY if hasattr(settings, 'ENCRYPT_KEY') else settings.SECRET_KEY
ENCRYPT_PRIVATE_KEY = settings.ENCRYPT_PRIVATE_KEY if hasattr(settings, 'ENCRYPT_PRIVATE_KEY') else None
ENCRYPT_PUBLIC_KEY = settings.ENCRYPT_PUBLIC_KEY if hasattr(settings, 'ENCRYPT_PUBLIC_KEY') else None



from django.db import connections, connection
from django.db.utils import InternalError
from django.core.management.base import BaseCommand

from sage_encrypt.services.setting import get_setting


class Command(BaseCommand):
    help = 'Decrypt all encrypted data in specific column.'

    def add_arguments(self, parser):
        """
        add arguments for command
        """
        parser.add_argument(
            '-d',
            '--database',
            type=str,
            help='database',
            required=False
        )
        parser.add_argument(
            '-t',
            '--table',
            type=str,
            help='database table',
            required=True
        )
        parser.add_argument(
            '-c',
            '--column',
            type=str,
            help='table column',
            required=True
        )

    def create_sym_query(self, table_name: str, column_name: str):
        """
        create symmetric decrypt & replace sql query
        :param table_name: target table name
        :type table_name: str
        :param column_name: target column name
        :type column_name: str
        :return: SQL Query
        :rtype: str
        """
        ENCRYPT_KEY = get_setting(connection, 'ENCRYPT_KEY')
        query = """
                UPDATE {table_name}
                    SET {col_name} = pgp_sym_decrypt({col_name}::bytea, '{sym_key}');
                """.format(
            table_name=table_name,
            col_name=column_name,
            sym_key=ENCRYPT_KEY
        )

        return query

    def create_asym_query(self, table_name: str, column_name: str):
        """
        create asymmetric decrypt & replace sql query
        :param table_name: target table name
        :type table_name: str
        :param column_name: target column name
        :type column_name: str
        :return: SQL Query
        :rtype: str
        """
        ENCRYPT_PRIVATE_KEY = get_setting(connection, 'ENCRYPT_PRIVATE_KEY')
        query = """
                UPDATE {table_name}
                    SET {col_name} = pgp_pub_decrypt({col_name}::bytea, dearmor('{asym_key}'));
                """.format(
            table_name=table_name,
            col_name=column_name,
            asym_key=ENCRYPT_PRIVATE_KEY
        )

        return query

    def handle(self, *args, **options):
        """
        execute raw sql to replace encrypted data
        """
        db_name = options['database']
        table_name = options['table']
        col_name = options['column']

        sym_sql_query = self.create_sym_query(table_name=table_name, column_name=col_name)
        asym_sql_query = self.create_asym_query(table_name=table_name, column_name=col_name)

        if db_name:
            with connections[db_name].cursor() as cursor:
                try:
                    cursor.execute(sym_sql_query)
                    message = f'Decryption applied.\nTable -> {table_name}\nColumn -> {col_name}\nAlgorithm -> symmetric'
                except InternalError:
                    cursor.execute(asym_sql_query)
                    message = f'Decryption applied.\nTable -> {table_name}\nColumn -> {col_name}\nAlgorithm -> asymmetric'
        else:
            with connection.cursor() as cursor:
                try:
                    cursor.execute(sym_sql_query)
                    message = f'Decryption applied.\nTable -> {table_name}\nColumn -> {col_name}\nAlgorithm -> symmetric'
                except InternalError:
                    cursor.execute(asym_sql_query)
                    message = f'Decryption applied.\nTable -> {table_name}\nColumn -> {col_name}\nAlgorithm -> asymmetric'

        self.stdout.write(
            self.style.SUCCESS(message)
        )



import logging

from django.db import models

from sage_encrypt.collections.asymmetric import ASYMMETRIC_FIELDS
from sage_encrypt.collections.symmetric import SYMMETRIC_FIELDS


def get_encrypted_field(base_class: models.Field, algorithm='symmetric'):
    """
    get encrypted field for given django field
    :type base_class: models.Field(*args, **kwargs)
    :type algorithm: str
    :rtype: EncryptedField(*args, **kwargs)
    """

    encrypt_fields = {
        'symmetric': SYMMETRIC_FIELDS,
        'asymmetric': ASYMMETRIC_FIELDS
    }

    assert isinstance(base_class, models.Field)
    name, path, args, kwargs = base_class.deconstruct()
    encrypted_field = encrypt_fields.get(algorithm).get(path)

    if encrypted_field:
        if name:
            base_class.__class__ = encrypted_field
            return base_class
        else:
            return encrypted_field(*args, **kwargs)
    else:
        logging.warning(
            'field type {} can not be encrypted in database, ignored'.format(path)
        )
        return base_class


def encrypt_field(base_field: models.Field, algorithm='symmetric'):
    """
    create EncryptedField for given Field
    :type base_field: models.Field(*args, **kwargs)
    :type algorithm: str (symmetric/asymmetric)
    :rtype: EncryptedField(*args, **kwargs)
    """
    return get_encrypted_field(base_field, algorithm)



from django.db import models

from sage_encrypt.mixins.cast import INT, BIGINT, DATE, TIMESTAMP, BOOL, DOUBLE, TIME, NUMERIC
from sage_encrypt.mixins.encrypt import EncryptAsymmetricMixin
from sage_encrypt.mixins.query import ASYM_ENCRYPT_SQL_WITH_NULLIF


class EncryptedEmailField(EncryptAsymmetricMixin, models.EmailField):
    pass


class EncryptedIntegerField(EncryptAsymmetricMixin, models.IntegerField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = INT

    def get_internal_type(self):
        return 'IntegerField'


class EncryptedBigIntegerField(EncryptAsymmetricMixin, models.BigIntegerField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = BIGINT

    def get_internal_type(self):
        return 'BigIntegerField'


class EncryptedTextField(EncryptAsymmetricMixin, models.TextField):
    pass


class EncryptedCharField(EncryptAsymmetricMixin, models.CharField):
    pass


class EncryptedDateField(EncryptAsymmetricMixin, models.DateField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = DATE

    def get_internal_type(self):
        return 'DateField'


class EncryptedDateTimeField(EncryptAsymmetricMixin, models.DateTimeField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = TIMESTAMP

    def get_internal_type(self):
        return 'DateTimeField'


class EncryptedBooleanField(EncryptAsymmetricMixin, models.BooleanField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = BOOL

    def get_internal_type(self):
        return 'BooleanField'


class EncryptedFloatField(EncryptAsymmetricMixin, models.FloatField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = DOUBLE

    def get_internal_type(self):
        return 'FloatField'


class EncryptedTimeField(EncryptAsymmetricMixin, models.TimeField):
    encrypt_query = ASYM_ENCRYPT_SQL_WITH_NULLIF
    cast = TIME

    def get_internal_type(self):
        return 'TimeField'

class EncryptedDecimalField(EncryptAsymmetricMixin, models.DecimalField):
    cast = NUMERIC

    def get_cast_sql(self):
        return self.cast % {
            'max_digits': self.max_digits,
            'decimal_places': self.decimal_places
        }

    def get_internal_type(self):
        return 'DecimalField'



from sage_encrypt.fields.symmetric import (
    EncryptedCharField,
    EncryptedEmailField,
    EncryptedDateField,
    EncryptedTextField,
    EncryptedFloatField,
    EncryptedDateTimeField,
    EncryptedDecimalField,
    EncryptedBooleanField,
    EncryptedIntegerField,
    EncryptedTimeField,
    EncryptedBigIntegerField
)

SYMMETRIC_FIELDS = {
    'django.db.models.CharField': EncryptedCharField,
    'django.db.models.DateField': EncryptedDateField,
    'django.db.models.DateTimeField': EncryptedDateTimeField,
    'django.db.models.IntegerField': EncryptedIntegerField,
    'django.db.models.BigIntegerField': EncryptedBigIntegerField,
    'django.db.models.EmailField': EncryptedEmailField,
    'django.db.models.TextField': EncryptedTextField,
    'django.db.models.FloatField': EncryptedFloatField,
    'django.db.models.DecimalField': EncryptedDecimalField,
    'django.db.models.BooleanField': EncryptedBooleanField,
    'django.db.models.TimeField': EncryptedTimeField
}



from setuptools import setup

setup(
    name='django-sage-encrypt',
    packages=['sage_encrypt'],
    version='0.4.4',
    license='GNU',
    description='encrypt PostgreSQL database',
    author='Sage Team',
    author_email='mail@sageteam.org',
    url='https://github.com/sageteam-org/django-sage-encrypt',
    download_url='https://github.com/sageteam-org/django-sage-encrypt/archive/refs/tags/0.1.tar.gz',
    keywords=['django', 'python', 'encrypt', 'PostgreSQL', 'postgres'],
    install_requires=[
        'Django',
        'psycopg2-binary'
    ]
)
